package latin2.computation

import info.kwarc.mmt.api.{objects, _}
import objects.{Stack, _}
import info.kwarc.mmt.api.execution._
import info.kwarc.mmt.api.frontend.Controller
import info.kwarc.mmt.api.symbols.Constant
import info.kwarc.mmt.lf._
import lf._


/**
 * Rule that runs before the execution process starts
 * This generates (recursive) rules from axiomatic declarations with type
 * axiom_name : functionName ↦ definiens
 */
object DefinedRun extends RulePreprocessor(RecurseableDefinitions.define.path){
  override def under: List[GlobalName]= List(Apply.path)
  def apply(const : Constant) : Option[ExecutionRule] = {
    const.tp match {
      case Some(RecurseableDefinitions.define(tp,OMID(gname: GlobalName),target : OMBINDC)) => {
        Some(new DefinedRule(gname, target))

      }
    }
  }
}

/**
 * Rule instance generated by DefinedRun
 */
class DefinedRule(override val head: GlobalName, val targetTerm : OMBINDC) extends ExecutionRule(head){
  override def under: List[GlobalName]= List(Apply.path)
  override def apply(controller:Controller, callback: ExecutionCallback, env: RuntimeEnvironment, prog: Term) : Term = {
    prog match {
      case ApplyGeneral(OMID(head), ls) => {
        val lsE = ls map callback.execute
        val app = ApplyGeneral(targetTerm,lsE)
        callback.execute(app)
      }
    }
  }
}
